# pylint: skip-file
# ruff: noqa

import logging
import os
from time import sleep

import pytest
from botocore.exceptions import ClientError
from opentaskpy.exceptions import CachingPluginError

from opentaskpy.variablecaching.aws import vc_secretsmanager
from tests.fixtures.localstack import *  # noqa: F403


def test_cacheable_variable_ssm_args():
    # Test combinations of the plugin with invalid args
    with pytest.raises(CachingPluginError):
        vc_secretsmanager.run()

    with pytest.raises(CachingPluginError):
        vc_secretsmanager.run(name="/test/variable")

    with pytest.raises(CachingPluginError):
        vc_secretsmanager.run(value="newvalue")


def test_cacheable_variable_vc_secretsmanager(secretsmanager_client):

    # Create a new ParamStore value
    secretsmanager_client.create_secret(
        Name="/test/variable",
        SecretString="originalvalue",
    )

    secretsmanager_client.get_secret_value(SecretId="/test/variable")

    assert (
        secretsmanager_client.get_secret_value(SecretId="/test/variable")[
            "SecretString"
        ]
        == "originalvalue"
    )

    kwargs = {"name": "/test/variable", "value": "newvalue"}
    vc_secretsmanager.run(**kwargs)

    # Check the variable has content of "newvalue"
    assert (
        secretsmanager_client.get_secret_value(SecretId="/test/variable")[
            "SecretString"
        ]
        == "newvalue"
    )


def test_cacheable_variable_secretsmanager_failure(secretsmanager_client):

    kwargs = {"name": "xxxx", "value": "newvalue"}

    # Remove the AWS_ENDPOINT_URL env var, so it tries to go to something that doesn't exist
    del os.environ["AWS_ENDPOINT_URL"]

    with pytest.raises(ClientError):
        vc_secretsmanager.run(**kwargs)


def test_cacheable_variable_vc_secretsmanager_min_cache_age(
    secretsmanager_client, caplog
):

    # Create a new ParamStore value
    secretsmanager_client.create_secret(
        Name="/test/min_cache_variable",
        SecretString="originalvalue",
    )

    secretsmanager_client.get_secret_value(SecretId="/test/min_cache_variable")

    assert (
        secretsmanager_client.get_secret_value(SecretId="/test/min_cache_variable")[
            "SecretString"
        ]
        == "originalvalue"
    )
    # Attempt update prior to min_cache_age passing since last update
    kwargs = {
        "name": "/test/min_cache_variable",
        "value": "newvalue",
        "min_cache_age": "60",
    }
    with caplog.at_level(logging.WARNING):
        vc_secretsmanager.run(**kwargs)
        assert "Not updating secret because secret last updated at" in caplog.text
    # Check the variable has not been updated
    assert (
        secretsmanager_client.get_secret_value(SecretId="/test/min_cache_variable")[
            "SecretString"
        ]
        == "originalvalue"
    )

    sleep(5)
    # Attempt update after min_cache_age has passed since last update
    kwargs = {
        "name": "/test/min_cache_variable",
        "value": "newvalue",
        "min_cache_age": "5",
    }
    vc_secretsmanager.run(**kwargs)

    # Check the variable has content of "newvalue"
    assert (
        secretsmanager_client.get_secret_value(SecretId="/test/min_cache_variable")[
            "SecretString"
        ]
        == "newvalue"
    )



"""AWS helper functions."""

import os
import socket
from time import time

import boto3
import opentaskpy.otflogging
from botocore.args import ClientArgsCreator
from botocore.config import Config

logger = opentaskpy.otflogging.init_logging(__name__, None, None)


def _custom_compute_socket_options(self, scoped_config, client_config=None):  # type: ignore[no-untyped-def]
    # This is a workaround for an issue in botocore - See the following PR for more details:
    # https://github.com/boto/botocore/pull/3140
    # Once this is merged, we can remove this monkey patch

    # This forces a more aggressive keepalive, as the default is too low for default
    # AWS NAT gateways where the idle timeout is set to 350 seconds

    socket_options = [(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)]
    client_keepalive = client_config and client_config.tcp_keepalive
    scoped_keepalive = (
        scoped_config
        and self._ensure_boolean(  # pylint: disable=protected-access
            scoped_config.get("tcp_keepalive", False)
        )
    )
    # Enables TCP Keepalive if specified in client config object or shared config file.
    if not client_keepalive and not scoped_keepalive:
        socket_options.append((socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1))
        return socket_options

    seconds_in_a_minute = 60

    client_read_timeout = client_config and client_config.read_timeout
    scoped_read_timeout = scoped_config.get("read_timeout", None)

    read_timeout = (
        scoped_read_timeout if scoped_read_timeout else client_read_timeout
    ) or seconds_in_a_minute

    maximum_keepalive_probes = int(read_timeout / seconds_in_a_minute) or 1
    keep_idle = (
        seconds_in_a_minute if read_timeout > seconds_in_a_minute else read_timeout
    )

    socket_options.extend(
        [
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, keep_idle),
            (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keep_idle),
            (
                socket.IPPROTO_TCP,
                socket.TCP_KEEPCNT,
                maximum_keepalive_probes,
            ),
        ],
    )
    return socket_options


def get_aws_client(
    client_type: str,
    credentials: dict,
    assume_role_arn: str | None = None,
    config: Config | None = None,
) -> dict:
    """Get an AWS client of the specified type using the provided credentials.

    Args:
        client_type: The type of client to get
        credentials: The credentials to use
        assume_role_arn: The role to assume
        config: The config to use for the client
    """
    if client_type == "lambda":
        # Monkey patch the socket options for lambda
        # See above for more details
        ClientArgsCreator._compute_socket_options = (  # pylint: disable=protected-access
            _custom_compute_socket_options
        )

    supported_types = ["s3", "ecs", "lambda", "logs"]
    if client_type not in supported_types:
        raise ValueError(
            f"Unsupported client type: {client_type}. Supported types are: {supported_types}"
        )

    kwargs = {}
    if os.environ.get("AWS_ENDPOINT_URL"):
        kwargs["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    if assume_role_arn:
        logger.info(f"Assuming role: {assume_role_arn}")
        sts_client = boto3.client("sts", **kwargs)

        assumed_role_object = sts_client.assume_role(
            RoleArn=assume_role_arn,
            RoleSessionName=f"OTF{time()}",
            DurationSeconds=900,
        )

        credentials = assumed_role_object["Credentials"]
        # Log the assumed role access key id
        logger.info(f"Assumed role access key id: {credentials['AccessKeyId']}")

    kwargs2 = {
        "aws_access_key_id": credentials["AccessKeyId"],
        "aws_secret_access_key": credentials["SecretAccessKey"],
    }
    if "SessionToken" in credentials:
        kwargs2["aws_session_token"] = credentials["SessionToken"]

    if "region_name" in credentials:
        kwargs2["region_name"] = credentials["region_name"]

    session = boto3.session.Session(**kwargs2)

    return {
        "client": session.client(client_type, **kwargs, config=config),
        "temporary_creds": credentials if assume_role_arn else None,
    }


def set_aws_creds(obj) -> None:  # type: ignore[no-untyped-def]
    """Set AWS credentials for boto3.

    Args:
        obj: The object to set the credentials on

    """
    obj.aws_access_key_id = (
        obj.spec["protocol"]["access_key_id"]
        if "access_key_id" in obj.spec["protocol"]
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    obj.aws_secret_access_key = (
        obj.spec["protocol"]["secret_access_key"]
        if "secret_access_key" in obj.spec["protocol"]
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    obj.assume_role_arn = (
        obj.spec["protocol"]["assume_role_arn"]
        if "assume_role_arn" in obj.spec["protocol"]
        else os.environ.get("AWS_ROLE_ARN")
    )

    obj.region_name = (
        obj.spec["protocol"]["region_name"]
        if "region_name" in obj.spec["protocol"]
        else os.environ.get("AWS_REGION")
    )

    obj.bucket_owner_full_control = obj.spec["protocol"].get(
        "bucket_owner_full_control", True
    )



"""Secrets Manager lookup plugin.

Uses boto3 to pull out a secret from AWS Secrets Manager
This uses AWS authentication, either from the IAM role of the host,
by using environment variables, or variables in variables.json file
"""

import json
import os

import boto3
import opentaskpy.otflogging
from botocore.exceptions import ClientError
from jsonpath_ng import parse
from opentaskpy.exceptions import LookupPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

plugin_name = "secretsmanager"


def run(**kwargs):  # type: ignore[no-untyped-def]
    """Pull a secret from AWS Secrets Manager.

    Args:
        **kwargs: Expect a kwarg named name. This should be the secret name
        to obtain. The value should be a string.
          Optionally, a kwarg named value pointing at the JSON path, if the secret is to
            be parsed as JSON. The value returned will be the value at the path. The
            value should be a string.


    Raises:
        LookupPluginError: Returned if the kwarg 'name' is not provided, or if the JSONPath
            returns a value of the wrong type
        ClientError: Returned if anything fails during the lookup

    Returns:
        _type_: The value read from Parameter Store. If there is an issue obtaining the
            value, then "LOOKUP_FAILED" is returned, instead of causing the whole task
            to fail, unless the environment variable OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR
            is set to 1.
    """
    # Expect a kwarg named name
    expected_kwargs = ["name"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise LookupPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run lookup plugin"
                f" '{plugin_name}'"
            )

    fail_on_exception = (
        os.environ.get("OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR", "0") == "1"
    )

    globals_ = kwargs.get("globals", None)

    aws_access_key_id = (
        globals_["AWS_ACCESS_KEY_ID"]
        if globals_ and "AWS_ACCESS_KEY_ID" in globals_
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key = (
        globals_["AWS_SECRET_ACCESS_KEY"]
        if globals_ and "AWS_SECRET_ACCESS_KEY" in globals_
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    region_name = (
        globals_["AWS_REGION"]
        if globals_ and "AWS_REGION" in globals_
        else os.environ.get("AWS_REGION")
    )
    boto3_kwargs = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "region_name": region_name,
    }
    # If there's an override for endpoint_url in the environment, then use that
    kwargs2 = {}
    if os.environ.get("AWS_ENDPOINT_URL"):
        kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    result = None
    try:
        session = boto3.session.Session(**boto3_kwargs)
        secretsmanager = session.client("secretsmanager", **kwargs2)
        response = secretsmanager.get_secret_value(SecretId=kwargs["name"])
        result = response["SecretString"]

        # Very simple redacting filter here for secrets, e.g. pgp private keys,
        # or ssh private keys
        log_result = result
        if " private " in result.lower():
            log_result = "REDACTED"

        logger.log(12, f"Read '{log_result}' from param {kwargs['name']}")

        # If requested to pull a value from the JSON, then parse it and extract the
        # value at the path
        if "value" in kwargs:
            try:
                result = json.loads(result)
                # Handle the JSONPath
                jsonpath_expr = parse(kwargs["value"])
                result = jsonpath_expr.find(result)

                result = result[0].value

                # If the result is a list, then return the first element, with a warning
                # that there's more than one
                if isinstance(result, list) and isinstance(result[0], (str, int)):
                    logger.warning(
                        f"JSONPath returned a list of length {len(result)}. Returning "
                        + "only the first element"
                    )
                    result = result[0]

                # If the result is not a string or an int, then raise an exception
                if not isinstance(result, (str, int)):
                    if fail_on_exception:
                        raise LookupPluginError(
                            f"JSONPath returned a value of type {type(result)}. Expected a string or int"
                        )
                    logger.warning(
                        f"JSONPath returned a value of type {type(result)}. Continuing anyway."
                    )
                    return "LOOKUP_FAILED"

            except json.JSONDecodeError as e:
                logger.error(f"Failed to parse JSON from secret: {kwargs['name']}: {e}")

                if fail_on_exception:
                    raise e

                logger.warning("Secret JSON parsing failed but continuing anyway")
                return "LOOKUP_FAILED"
            except KeyError as e:
                logger.error(f"Failed to read from secret: {kwargs['name']}: {e}")

                if fail_on_exception:
                    raise e

                logger.warning("Secret JSON parsing failed but continuing anyway")
                return "LOOKUP_FAILED"

    except ClientError as e:
        # To prevent complete failure of all jobs in an environment on failed lookup
        # return 'LOOKUP_FAILED and log error instead of throwing exception'
        if e.response["Error"]["Code"] == "ResourceNotFoundException":
            logger.error(f"Secret not found: {kwargs['name']}: {e}")
        else:
            logger.error(f"Failed to read from Secrets Manager: {kwargs['name']}: {e}")

        if fail_on_exception:
            raise e

        logger.warning("Secrets Manager lookup failed but continuing anyway")
        return "LOOKUP_FAILED"
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f"Failed to read from Secrets Manager: {kwargs['name']}: {e}")

        if fail_on_exception:
            raise e

        logger.warning("Secrets Manager lookup failed but continuing anyway")
        return "LOOKUP_FAILED"

    # Escape any escape characters so they can be stored in JSON as a string
    if result and isinstance(result, str):
        # Escape any newline characters
        result = result.replace("\n", "\\n")
        result = json.dumps(result)
        # Remove the leading and trailing quotes
        result = result[1:-1]

    return result



# pylint: skip-file
# ruff: noqa
import json
import logging
from json.decoder import JSONDecodeError

import opentaskpy.otflogging
import pytest
from botocore.exceptions import ClientError
from opentaskpy.config.loader import ConfigLoader
from opentaskpy.exceptions import LookupPluginError
from pytest_shell import fs

from opentaskpy.plugins.lookup.aws.secrets_manager import run
from tests.fixtures.localstack import *  # noqa: F403

PLUGIN_NAME = "secretsmanager"

# Get the default logger and set it to DEBUG
logger = logging.getLogger()
logger.setLevel(12)


def test_secrets_manager_plugin_missing_name():
    with pytest.raises(Exception) as ex:
        run()

    assert (
        ex.value.args[0]
        == f"Missing kwarg: 'name' while trying to run lookup plugin '{PLUGIN_NAME}'"
    )


def test_secrets_manager_plugin_secret_not_found(credentials, caplog):
    # Ensure lookups throw an exception when lookups fail

    os.environ["OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR"] = "1"

    with pytest.raises(ClientError) as ex:
        result = run(name="does_not_exist")

    # Ensure secrets lookup failures return "LOOKUP_FAILED" and log warning rather than raising an exception to prevent full environment failure,
    # when OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR env var is not set

    del os.environ["OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR"]

    with caplog.at_level(logging.WARNING):
        result = run(name="does_not_exist")
    assert "Secrets Manager lookup failed but continuing anyway" in caplog.text
    assert result == "LOOKUP_FAILED"


def test_secrets_manager_plugin_standard_string(secrets_manager_client):
    expected = "test1234"
    # Populate a Secrets Manager secret with a test value
    secrets_manager_client.create_secret(Name="standard_string", SecretString=expected)

    result = run(name="standard_string")

    assert result == expected


def test_secrets_manager_plugin_redacted_secret(secrets_manager_client, caplog):
    expected = " private "
    # Populate a Secrets Manager secret with a test value
    secrets_manager_client.create_secret(
        Name="redacted_secret", SecretString=" private "
    )

    with caplog.at_level(12):
        # Ensure the logger is set to the same VERBOSE level
        opentaskpy.otflogging.init_logging(
            "opentaskpy.plugins.lookup.aws.secrets_manager"
        )
        result = run(name="redacted_secret")
    assert "Read 'REDACTED' from param redacted_secret" in caplog.text

    assert result == expected


def test_config_loader_using_secretsmanager_plugin(secrets_manager_client, tmpdir):
    json_obj = {
        "testLookup": "{{ lookup('aws.secrets_manager', name='standard_string_multiline') }}",
    }

    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )
    # Test with a multi line string to make sure that it doesn't break the parser
    expected_result = """config_loader_test_1234\\nanother_line"""

    # Insert the param into secrets manager
    secrets_manager_client.create_secret(
        Name="standard_string_multiline", SecretString=expected_result
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)
    config_loader._load_global_variables()
    config_loader._resolve_templated_variables()

    assert config_loader.get_global_variables()["testLookup"] == expected_result


def test_config_loader_using_secretsmanager_plugin_json_path(
    secrets_manager_client, tmpdir
):
    json_obj = {
        "testLookup": "{{ lookup('aws.secrets_manager', name='plugin_json_path', value='foo.bar.[1]') }}",
    }

    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )
    # Test with a multi line string to make sure that it doesn't break the parser
    expected_result = "nested_secret_value"

    # Insert the param into paramstore
    secrets_manager_client.create_secret(
        Name="plugin_json_path",
        SecretString=json.dumps(
            {"foo": {"bar": ["first_secret_value", "nested_secret_value"]}}
        ),
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)
    config_loader._load_global_variables()
    config_loader._resolve_templated_variables()

    assert config_loader.get_global_variables()["testLookup"] == expected_result


def test_config_loader_using_secretsmanager_plugin_invalid_return_types(
    secrets_manager_client, tmpdir, caplog
):
    # Insert the param into paramstore
    secrets_manager_client.create_secret(
        Name="invalid_return_types",
        SecretString=json.dumps(
            {"foo": {"bar": ["first_secret_value", "nested_secret_value"]}}
        ),
    )

    json_obj = {
        "testLookup": "{{ lookup('aws.secrets_manager', name='invalid_return_types', value='foo') }}",
    }

    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )
    # Test that the global variables are loaded correctly
    with caplog.at_level(logging.WARNING):
        config_loader = ConfigLoader(tmpdir)
        config_loader._load_global_variables()
        config_loader._resolve_templated_variables()

    assert (
        "JSONPath returned a value of type <class 'dict'>. Continuing anyway."
        in caplog.text
    )
    assert config_loader.get_global_variables()["testLookup"] == "LOOKUP_FAILED"

    # Run again but with exceptions being raised
    os.environ["OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR"] = "1"
    with pytest.raises(LookupPluginError) as ex:
        config_loader = ConfigLoader(tmpdir)
        config_loader._load_global_variables()
        config_loader._resolve_templated_variables()

    del os.environ["OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR"]

    # Delete the previous variables file
    os.remove(f"{tmpdir}/variables.json.j2")

    # Now do a list
    json_obj = {
        "testLookup": "{{ lookup('aws.secrets_manager', name='invalid_return_types', value='foo.bar') }}",
    }
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )

    with caplog.at_level(logging.WARNING):
        config_loader = ConfigLoader(tmpdir)
        config_loader._load_global_variables()
        config_loader._resolve_templated_variables()

    assert (
        "JSONPath returned a list of length 2. Returning only the first element"
        in caplog.text
    )
    assert config_loader.get_global_variables()["testLookup"] == "first_secret_value"


def test_config_loader_using_secretsmanager_plugin_invalid_json(
    secrets_manager_client, tmpdir, caplog
):
    # Insert the param into paramstore
    secrets_manager_client.create_secret(
        Name="invalid_return_json",
        SecretString=json.dumps(
            {"foo": {"bar": ["first_secret_value", "nested_secret_value"]}}
        )
        + "}}}}",  # Intentionally break the JSON
    )

    json_obj = {
        "testLookup": "{{ lookup('aws.secrets_manager', name='invalid_return_json', value='foo') }}",
    }

    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )
    # Test that the global variables are loaded correctly
    with caplog.at_level(logging.WARNING):
        config_loader = ConfigLoader(tmpdir)
        config_loader._load_global_variables()
        config_loader._resolve_templated_variables()

    assert "Secret JSON parsing failed but continuing anyway" in caplog.text
    assert config_loader.get_global_variables()["testLookup"] == "LOOKUP_FAILED"

    # Run again but with exceptions being raised
    os.environ["OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR"] = "1"
    with pytest.raises(JSONDecodeError) as ex:
        config_loader = ConfigLoader(tmpdir)
        config_loader._load_global_variables()
        config_loader._resolve_templated_variables()



# pylint: skip-file
# ruff: noqa

import os

import pytest
from botocore.exceptions import ClientError
from opentaskpy.exceptions import CachingPluginError

from opentaskpy.variablecaching.aws import vc_ssm
from tests.fixtures.localstack import *  # noqa: F403


def test_cacheable_variable_ssm_args():
    # Test combinations of the plugin with invalid args
    with pytest.raises(CachingPluginError):
        vc_ssm.run()

    with pytest.raises(CachingPluginError):
        vc_ssm.run(name="/test/variable")

    with pytest.raises(CachingPluginError):
        vc_ssm.run(value="newvalue")


def test_cacheable_variable_ssm(ssm_client):

    # Create a new ParamStore value
    ssm_client.put_parameter(
        Name="/test/variable",
        Value="originalvalue",
        Type="SecureString",
        Overwrite=True,
    )

    param = ssm_client.get_parameter(Name="/test/variable", WithDecryption=True)

    spec = {"task_id": "1234", "x": {"y": "value"}}

    kwargs = {"name": f"/test/variable", "value": "newvalue"}

    vc_ssm.run(**kwargs)

    # Check the variable has content of "newvalue"
    assert (
        ssm_client.get_parameter(Name="/test/variable", WithDecryption=True)[
            "Parameter"
        ]["Value"]
        == "newvalue"
    )


def test_cacheable_variable_ssm_failure(ssm_client):

    spec = {"task_id": "1234", "x": {"y": "value"}}

    kwargs = {"name": f"xxxx", "value": "newvalue"}

    # Remove the AWS_ENDPOINT_URL env var, so it tries to go to something that doesn't exist
    del os.environ["AWS_ENDPOINT_URL"]

    with pytest.raises(ClientError):
        vc_ssm.run(**kwargs)



"""Caching plugin for writing variables to AWS SSM."""

import os

import boto3
import opentaskpy.otflogging
from botocore.exceptions import ClientError
from opentaskpy.exceptions import CachingPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

CACHE_NAME = "vc_ssm"


def run(**kwargs):  # type: ignore[no-untyped-def]
    """Write a variable to AWS SSM.

    Args:
        **kwargs: Expect kwargs named 'name', and 'value'. This should be the Parameter
         Store parameter value to write to, and the value to put into the file

    Raises:
        CachingPluginError: Returned if the kwarg 'name' or 'value' is not provided
        FileNotFoundError: Returned if the file does not exist
    """
    # Expect a kwarg named name, and value
    expected_kwargs = ["name", "value"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise CachingPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run caching plugin"
                f" '{CACHE_NAME}'"
            )

    globals_ = kwargs.get("globals", None)

    aws_access_key_id = (
        globals_["AWS_ACCESS_KEY_ID"]
        if globals_ and "AWS_ACCESS_KEY_ID" in globals_
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key = (
        globals_["AWS_SECRET_ACCESS_KEY"]
        if globals_ and "AWS_SECRET_ACCESS_KEY" in globals_
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    region_name = (
        globals_["AWS_REGION"]
        if globals_ and "AWS_REGION" in globals_
        else os.environ.get("AWS_REGION")
    )
    boto3_kwargs = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "region_name": region_name,
    }
    # If there's an override for endpoint_url in the environment, then use that
    kwargs2 = {}
    if os.environ.get("AWS_ENDPOINT_URL"):
        kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    try:
        session = boto3.session.Session(**boto3_kwargs)
        ssm = session.client("ssm", **kwargs2)

        # Write the value to the SSM parameter
        ssm.put_parameter(
            Name=kwargs["name"],
            Value=kwargs["value"],
            Type="SecureString",
            Overwrite=True,
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "ParameterNotFound":
            logger.error(f"Parameter not found: {kwargs['name']}: {e}")
            raise e
        raise e
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f"Failed to write SSM parameter: {kwargs['name']}: {e}")
        raise e



"""SSM Param Store lookup plugin.

Uses boto3 to pull out a parameter from AWS Parameter Store
This uses AWS authentication, either from the IAM role of the host,
by using environment variables, or variables in variables.json file
"""

import json
import os

import boto3
import opentaskpy.otflogging
from botocore.exceptions import ClientError
from opentaskpy.exceptions import LookupPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

plugin_name = "ssm"


def run(**kwargs):  # type: ignore[no-untyped-def]
    """Pull a variable from AWS SSM.

    Args:
        **kwargs: Expect a kwarg named name. This should be the key within SSM to the
        variable to obtain. The value should be a string.

    Raises:
        LookupPluginError: Returned if the kwarg 'name' is not provided
        ClientError: Returned if the parameter does not exist

    Returns:
        _type_: The value read from Parameter Store
    """
    # Expect a kwarg named name
    expected_kwargs = ["name"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise LookupPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run lookup plugin"
                f" '{plugin_name}'"
            )

    fail_on_exception = (
        os.environ.get("OTF_AWS_SECRETS_LOOKUP_FAILED_IS_ERROR", "0") == "1"
    )

    globals_ = kwargs.get("globals", None)

    aws_access_key_id = (
        globals_["AWS_ACCESS_KEY_ID"]
        if globals_ and "AWS_ACCESS_KEY_ID" in globals_
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key = (
        globals_["AWS_SECRET_ACCESS_KEY"]
        if globals_ and "AWS_SECRET_ACCESS_KEY" in globals_
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    region_name = (
        globals_["AWS_REGION"]
        if globals_ and "AWS_REGION" in globals_
        else os.environ.get("AWS_REGION")
    )
    boto3_kwargs = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "region_name": region_name,
    }
    # If there's an override for endpoint_url in the environment, then use that
    kwargs2 = {}
    if os.environ.get("AWS_ENDPOINT_URL"):
        kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    result = None
    try:
        session = boto3.session.Session(**boto3_kwargs)
        ssm = session.client("ssm", **kwargs2)
        response = ssm.get_parameter(Name=kwargs["name"], WithDecryption=True)
        result = response["Parameter"]["Value"]

        # Very simple redacting filter here for secrets, e.g. pgp private keys,
        # or ssh private keys
        log_result = result
        if " private " in result.lower():
            log_result = "REDACTED"

        logger.log(12, f"Read '{log_result}' from param {kwargs['name']}")

    except ClientError as e:
        # To prevent complete failure of all jobs in an environment on failed lookup return 'LOOKUP_FAILED and log error instead of throwing exception'
        if e.response["Error"]["Code"] == "ParameterNotFound":
            logger.error(f"Parameter not found: {kwargs['name']}: {e}")
        else:
            logger.error(f"Failed to read from SSM parameter: {kwargs['name']}: {e}")

        if fail_on_exception:
            raise e

        logger.warning("SSM parameter lookup failed but continuing anyway")
        return "LOOKUP_FAILED"
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f"Failed to read from SSM parameter: {kwargs['name']}: {e}")

        if fail_on_exception:
            raise e

        logger.warning("SSM parameter lookup failed but continuing anyway")
        return "LOOKUP_FAILED"

    # Escape any escape characters so they can be stored in JSON as a string
    if result:
        # Escape any newline characters
        result = result.replace("\n", "\\n")
        result = json.dumps(result)
        # Remove the leading and trailing quotes
        result = result[1:-1]

    return result



# pylint: skip-file
# ruff: noqa

from opentaskpy.taskhandlers import transfer

from opentaskpy.variablecaching.aws import vc_ssm
from tests.fixtures.localstack import *  # noqa: F403

dummy_task_definition = {
    "task_id": 1234,
    "type": "transfer",
    "source": {
        "accessToken": "0",
        "protocol": {"name": "dummy"},
        "cacheableVariables": [
            {
                "variableName": "accessToken",
                "cachingPlugin": "aws.vc_ssm",
                "cacheArgs": {
                    "name": "/test/variablename",
                },
            }
        ],
    },
}


def test_dummy_transfer(ssm_client):
    #  The key thing to test is that the access token
    #  is written to the cache file

    transfer_obj = transfer.Transfer(None, "dummy_task_transfer", dummy_task_definition)

    transfer_obj.run()

    # Check the parameter store value now exists and contains a random number
    param = ssm_client.get_parameter(Name="/test/variablename", WithDecryption=True)
    assert param["Parameter"]["Value"] != "0"
    assert param["Parameter"]["Value"].isdigit()



"""Caching plugin for writing variables to AWS Secrets Manager."""

import os
from datetime import datetime, timedelta

import boto3
import opentaskpy.otflogging
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal
from opentaskpy.exceptions import CachingPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

CACHE_NAME = "vc_secretsmanager"


def run(**kwargs):  # type: ignore[no-untyped-def]
    """Write a variable to AWS Secrets Manager.

    Args:
        **kwargs: Expect kwargs named 'name', and 'value'. This should be the Secrets
         Manager secret to write to, and the value to put into the file.
         Optional kwargs named 'min_cache_age'. This should be the minimum amount of time (in seconds) to wait before saving this variable in secrets manager again
         (recommended > 10 mins to avoid possible issues with running out of secret versions)

    Raises:
        CachingPluginError: Returned if the kwarg 'name' or 'value' is not provided
        FileNotFoundError: Returned if the file does not exist
    """
    # Expect a kwarg named name, and value
    expected_kwargs = ["name", "value"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise CachingPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run caching plugin"
                f" '{CACHE_NAME}'"
            )

    globals_ = kwargs.get("globals", None)

    aws_access_key_id = (
        globals_["AWS_ACCESS_KEY_ID"]
        if globals_ and "AWS_ACCESS_KEY_ID" in globals_
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key = (
        globals_["AWS_SECRET_ACCESS_KEY"]
        if globals_ and "AWS_SECRET_ACCESS_KEY" in globals_
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    region_name = (
        globals_["AWS_REGION"]
        if globals_ and "AWS_REGION" in globals_
        else os.environ.get("AWS_REGION")
    )
    boto3_kwargs = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "region_name": region_name,
    }
    # If there's an override for endpoint_url in the environment, then use that
    kwargs2 = {}
    if os.environ.get("AWS_ENDPOINT_URL"):
        kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    try:
        session = boto3.session.Session(**boto3_kwargs)
        secrets_manager = session.client("secretsmanager", **kwargs2)

        # Check if this secret has been updated more recently than min cache age (if applicable)
        if "min_cache_age" in kwargs:
            secret = secrets_manager.describe_secret(
                SecretId=kwargs["name"],
            )
            if secret["LastChangedDate"] > datetime.now(tz=tzlocal()) - timedelta(
                seconds=int(kwargs["min_cache_age"])
            ):
                logger.warning(
                    f"Not updating secret because secret last updated at {secret['LastChangedDate']}, and min_cache_age is {kwargs['min_cache_age']}"
                )
                return

        # Write the value to secrets manager
        secrets_manager.put_secret_value(
            SecretId=kwargs["name"],
            SecretString=kwargs["value"],
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "ResourceNotFoundException":
            logger.error(f"Secret not found: {kwargs['name']}: {e}")
            raise e
        raise e
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f"Failed to write secret: {kwargs['name']}: {e}")
        raise e



"""AWS Lambda remote handler."""

import base64
import json
from datetime import datetime, timedelta
from typing import Any

import boto3
import opentaskpy.otflogging
from botocore.config import Config
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal
from opentaskpy.exceptions import InvalidConfigError
from opentaskpy.remotehandlers.remotehandler import RemoteExecutionHandler

from .creds import get_aws_client, set_aws_creds


class LambdaExecution(RemoteExecutionHandler):
    """AWS Lambda remote handler."""

    TASK_TYPE = "E"

    def tidy(self) -> None:
        """Tidy up the lambda client."""
        self.lambda_client.close()  # type: ignore[has-type]

    def __init__(self, spec: dict):
        """Initialise the LambdaExecution handler.

        Args:
            spec (dict): The spec for the execution.
        """
        # Ensure that function_arn is defined in the spec
        # This is really handled by the schema checks

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        self.aws_access_key_id: str | None = None
        self.aws_secret_access_key: str | None = None
        self.region_name: str | None = None

        super().__init__(spec)

        if "functionArn" not in self.spec:
            raise InvalidConfigError("functionArn not defined in spec")

        self.temporary_creds: dict | None = None
        self.assume_role_arn: str | None
        self.lambda_client: boto3.Client = None

        super().__init__(spec)

        set_aws_creds(self)

        self.credentials: dict = {
            "AccessKeyId": self.aws_access_key_id,
            "SecretAccessKey": self.aws_secret_access_key,
            "region_name": self.region_name,
        }

        self.validate_or_refresh_creds()

    def validate_or_refresh_creds(self) -> None:
        """Check the expiry of the temporary credentials, if applicable."""
        if self.lambda_client and not self.temporary_creds:
            return

        if self.temporary_creds:
            self.logger.debug(
                f"Temporary creds expire at: {self.temporary_creds['Expiration']} - Now: {datetime.now(tz=tzlocal())}"
            )

        if not self.lambda_client or (
            self.temporary_creds
            and self.temporary_creds["Expiration"]
            < datetime.now(tz=tzlocal()) + timedelta(minutes=1)
        ):

            if self.temporary_creds:
                self.logger.info("Renewing temporary credentials")

            # Set boto retries to 0 unless explicitly overridden in spec - retries will normally be handled within lambda code if required
            config_options: dict[str, Any] = {}
            if "max_attempts" in self.spec["protocol"]:
                config_options["retries"] = {
                    "max_attempts": self.spec["protocol"]["max_attempts"]
                }
                self.logger.info(
                    f"Setting max attempts to {self.spec['protocol']['max_attempts']}"
                )
            else:
                config_options["retries"] = {"max_attempts": 0}
            # If protocol has a botocoreReadTimeout set, then create a custom config with that set
            if "botocoreReadTimeout" in self.spec["protocol"]:
                config_options["read_timeout"] = self.spec["protocol"][
                    "botocoreReadTimeout"
                ]
                config_options["tcp_keepalive"] = True

            config = Config(**config_options)

            client_result = get_aws_client(
                "lambda", self.credentials, self.assume_role_arn, config=config
            )
            self.temporary_creds = (
                client_result["temporary_creds"]
                if client_result["temporary_creds"]
                else None
            )
            self.lambda_client = client_result["client"]

    def kill(self) -> None:
        """Kill the lambda function.

        This doesn't do a huge amount. Lambda functions don't run for a long period of
        time anyway, so this is more a way to clean up the client nicely.
        """
        # We need to kill the running execute function
        self.tidy()
        self.logger.info("Closed Lambda client")

    def execute(self) -> bool:
        """Execute the lambda function.

        Triggers the lambda function. The synchronous invocation type will cause this
        thread to block. This can be an issue with batch jobs and their timeout. Boto3
        will timeout after 60 seconds regardless, so Lambda functions that are called
        should not be long running. Their own timeout should be less than the batch timeout
        if one is used.

        An async call will not check the status of the lambda function, only if there
        are errors with invoking it.
        """
        result = True

        function_arn = self.spec["functionArn"]
        invocation_type = self.spec.get("invocationType", "Event")
        payload = None
        if "payload" in self.spec:
            payload = self.spec["payload"]

        try:
            invoke_response = self.lambda_client.invoke(
                FunctionName=function_arn,
                InvocationType=invocation_type,
                LogType="Tail",
                Payload=json.dumps(payload),
            )

            self.logger.info(f"Got status code: {invoke_response['StatusCode']}")
            # Print the response
            self.logger.info(f"Lambda function response: {invoke_response}")

            # Print the response payload
            self.logger.info(
                f"Got the following response from Lambda invocation: {invoke_response['Payload'].read().decode('utf-8')}"
            )

            if (
                invoke_response["StatusCode"] != 200
                and invocation_type == "RequestResponse"
            ) or (invoke_response["StatusCode"] != 202 and invocation_type == "Event"):
                self.logger.error(f"Failed to run lambda function: {function_arn}")
                return False

            # Log the response if there is one
            if "LogResult" in invoke_response:
                # base64 decode the result
                log_result = base64.b64decode(invoke_response["LogResult"]).decode()

                self.logger.info(f"Lambda function log: {log_result}")

            # Also see if there's any actual result body
            if "Payload" in invoke_response:
                result_payload = invoke_response["Payload"].read()
                self.logger.debug(f"Lambda function payload: {result_payload}")

            if "FunctionError" in invoke_response:
                self.logger.error(
                    f"Lambda function returned an error: {function_arn} - AWS Exception"
                    f" message: {invoke_response['FunctionError']}"
                )
                return False

        except ClientError as e:
            self.logger.error(f"Failed to run lambda function: {function_arn}")
            self.logger.error(e)
            result = False

        return result
